
/**
 * @file         BaseClass.js
 * @fileoverview Alternate Collection of routines to emulate multiple-inheritance
 * classes in JavaScript. This "multiple inheritance" has the earlier
 * classes in the superclass list getting precedence over the later classes
 * with regard to resolving both explicit and implicit-via-super references.
 * @requires     Validator.js (optional; only required if validating constructor args)
 */
/**
 * This function returns the name of a given function; It does this by
 * using a regular expression to extract the function name from the
 * function source code.
 * @param {Function} f reference to a function
 * @return name of given function
 * @type String
 */
function jsFuncName(f)
{
	var s = f.toString().match(/function\s*(\S*)\s*\(/)[1];
	return s ? s : "anonymous";
}

Class(GuiObject);
/**
 * @class This is the root class for all "classes" implemented
 * via the {@link GLOBALS#Class} function (ala Java "Object").
 * Class replaces the normal Javascript "constructor" function
 * with a new wrapper function that implements a robust
 * inheritence mechanism.
 * This mechanism supports "pretty" source code whereby
 * "methods" are declared inline, but the source code of
 * those methods are not copied into every object instance
 * as would otherwise be the case.<p>
 * The wrapper function also sets up a real "constructor"
 * that can use {@link GLOBALS#ValidateArgs} to insure
 * that the required parameters are passed to it, throwing
 * an exception if not.<p>
 * The design pattern implemented by Class/Extends corrects
 * several problems that occur with the naive inheritence
 * mechanisms often used with Javascript.
 *<pre>
 *  The required coding convention for "GuiObject":
 *  (1) the "class" is declared via the utility
 *      function {@link GLOBALS#Class}
 *  (2) the "superclass" (if any) is declared via the
 *      .Extends method {@link GLOBALS#Extends}
 *  (3) the "constructor" logic for the class is placed
 *      in a "method" named "konstructor"
 *  (4) the konstructor method invokes the "super" constructor
 *      via its class name [rather than literally "super()"]
 *  (5) "interfaces" and extra "superclasses" (if any) are
 *      declared via the .Implements method {@link GLOBALS#Implements}
 *</pre>
 * @throws MissingArgException thrown by constructor wrapper
 * @see GLOBALS#Class
 * @see GLOBALS#Extends
 * @see GLOBALS#Implements
 
 */
function GuiObject()
{
	/** Invokes the "super" version of whatever method that is calling superclass.
	 * @return either the return value of the super method or null if no super exists
	 */
	this.superclass = function() {
		var superMethod = this.superclass.caller.superclass;
		return (superMethod) ? superMethod.apply( this, arguments ) : null;
	}

	/** Invokes the version of whatever method that is calling superclass_ from
	 * the specified superclass.
	 * @return either the return value of the super method or null if no super exists
	 */
	this.superclass_ = function( superclass ) {
		var superMethod = this[superclass.cname]._[this.superclass_.caller.mname];
		var theArgs     = Array.prototype.slice.call(arguments,1);
		return (superMethod) ? superMethod.apply( this, theArgs ) : null;
	}

	/** Is "this" a descendent of the given class?.
	 * This implementation works with  multiple inheritence.
	 * @param {Function} aClass the class to verify
	 * @return true iff "this" is an "instanceof" aClass
	 */
	this.isInstanceOf = function( aClass ) {
		if (aClass.cname==null) return false;
		var m = this[ aClass.cname ];
		return( m && m instanceof Function );
	}
}


/**
 * This utility function sets up multiple "implements" inheritance
 * with all of the interfaces/classes passed in as parameters.
 * Multiple inheritance is implemented where method references
 * (either explicit or implicit-via-super) are resolved with
 * left-to-right precedence (with "Extends" superclass getting highest)
 * @see #Class
 * @see GuiObject
 */
function Implements()
{
	///////////////////////////////////////////////////////////////////
	// NOTE: At this point each superClass should have already had its
	// own Class and Extends calls performed. I.E. always "declare"
	// superclasses BEFORE any classes that extend them.
	///////////////////////////////////////////////////////////////////
	var N = arguments.length;
	for (var i=0; i<N; ++i)
	  if (!arguments[i].baseClass)
	    throw "Undeclared Super Class in .Implements";

	var s, m, superExemplar;
	var         rawExemplar = this.rawExemplar;
	var            exemplar = this.prototype;

	//foreach superclass, zip together "superclass" links
	//to each ORIGINAL subclass method not already superclass-linked.
	//
	//I.E. foreach superclass,
	//  foreach method in RAW exemplar,
	//    if method in CURRENT exemplar doesnt already have a superclass link
	//    then if method exists in current superclass exemplar
	//         then zip together "superclass" reference
	for (var i=0; i<N; ++i)
	{
		superExemplar = arguments[i].prototype;
		for (var property in rawExemplar)
		{
			if (!(rawExemplar[property] instanceof Function)) continue;
			m =      exemplar[property];  if (m.superclass) continue; //already linked!
			s = superExemplar[property];
			if (s && (s instanceof Function)) m.superclass = s;
		}
	}

	//foreach superclass, copy into exemplar a reference
	//to each superclass method not already in exemplar
	for (var i=0; i<N; ++i)
	{
		superExemplar = arguments[i].prototype;
		for (var property in superExemplar)
		{
			s = superExemplar[property];
			if ( exemplar[property]===void 0 && s instanceof Function )
			     exemplar[property] = s;
		}
	}
}


/**
 * This utility function sets up single inheritance.
 * @param {Function} superClass the "superclass constructor" function
 * @return the "class object" (an extended Function object)
 */
function Extends( superClass )
{
	///////////////////////////////////////////////////////////////////
	// NOTE: At this point superClass should have already had its own
	// Class and Extends calls performed. I.E. "declare" superclasses
	// BEFORE those classes that extend them.
	///////////////////////////////////////////////////////////////////
	if (!superClass.baseClass) throw "Undeclared Super Class in .Extends";

	//get superclass exemplar but avoid infinite recursion on implied root class
	var superexemplar = (this==GuiObject) ? null
					  : (superClass ? superClass.prototype : null);

	//Recall that one can't change the exemplar.__proto__ property directly
	//SO, one must create a new exemplar instance after baseClass.prototype is set.
	this.baseClass.prototype = superexemplar;
	var exemplar = new this.baseClass();

	//for each method in exemplar, if there is a different method with the
	//same name in superexemplar, setup the .superclass link between them.
	//To support multiple inheritence disambiguation, save the method name too.
	var s,m;
	for (var property in exemplar)// iterate over all properties
	{
		m =      exemplar[property]; if (!(m instanceof Function)) continue;
		m.mname =         property;  if (! superexemplar         ) continue;
		s = superexemplar[property];
		if (s && (s!=m) && (s instanceof Function)) m.superclass = s;
	}

	//Note that it would normally be a problem if Extends were called after
	//someone has already created an instance of "this" class (via new).
	//However, that situation should never occur since all the calls to Class()
	//and Extends() are normally done at web page load time, and application
	//logic should only start after the page "onload" event, hence all these
	//exemplars should be stable.
	this.prototype = exemplar;
	exemplar.name  = this.cname;//default value

	// while "superclass" works with konstructor, some may still prefer to
	// use the superclass name explicitly when invoking its constructor
	// so we leave this in to continue support for that.
	if (exemplar.konstructor==null) exemplar.konstructor = function(){};
	exemplar[this.cname]   = exemplar.konstructor;
	// and it is required for multiple-inheritence to resolve ambiguity
	exemplar.konstructor._ = exemplar;
	//save the original (unextended) exemplar to support .Implements
	if (superClass==GuiObject) this.rawExemplar = exemplar;

	return this;
}


/**
 * Declare a class and create the glue objects and code
 * to allow "pretty" source; A mechanism is set up to insure
 * that required constructor parameters each have a defined value;
 * This function also defines the "name" attribute of the specified
 * class and initializes it with the class name.
 * @param {Function} theClass the "class constructor" function
 * @param {StringArray} optConstructorArgDescArray optional array
 * of strings describing required parameters for this class'
 * "constructor"; Note: Optional constructor parameters should
 * not be included in this array.
 * @return the "class object" (an extended Function object)
 * @type Function
 * @see #Extends
 * @see #ValidateArgs
 * @see GuiObject
 */
function Class( theClass, optConstructorArgDescArray )
{
	
	// The "class" as passed in will really be an abstract base class. We
	// will squirrel that away and replace theClass with a wrapper function
	// so that method source code is not copied into each object instance
	// as would otherwise be the case.
	var baseClass = theClass;

	// create the new wrapper function to replace theClass
	var  className  = jsFuncName( theClass );
	self[className] = theClass = function(){
		var C = arguments.callee;
		if (C.required && self["ValidateArgs"]!==void 0)
		    ValidateArgs( C.cname, C.required, arguments );
		C.prototype.konstructor.apply( this, arguments );
	}
	theClass.required  = optConstructorArgDescArray;
	theClass.cname     = className; //FYI,netscape&firefox break with ".name"
	theClass.baseClass = baseClass;
	theClass.Extends   = Extends;//so we can call as methods...
	theClass.Implements= Implements;
	theClass.Extends(GuiObject);	//required here even if overridden later
	return theClass;
}